/********************

 SWIG-Goo!

 Andrew Sutherland
 January, 2002

 Currently heavily derived from the guile module for SWIG
  (and hopefully will stay that way... things look a lot
  uglier than the 1.1 docs make them out to be :)

 *******************/

#include "mod11.h"
#include "goo.h"
#include "swigconfig.h"
#include "swig.h"

#include <ctype.h>

static  File         *f_runtime = 0;
static  File         *f_header = 0;
static  File         *f_wrappers = 0;
static  File         *f_init = 0;

#ifdef SPLAT_MODULE_INFO
static  File         *f_binding_info = 0;
static  File         *f_export_info = 0;
#endif

static  File *f_goo_stub = NULL;
static  String *goo_text = NULL;

static char *stub_to_goo_use = NULL;
static char *stub_to_lib_use = NULL;

int dudes_so_far = 0;

GOO::GOO()
{
  SWIG_typemap_lang("goo");

  goo_text = NewString("(seq\n");
}

void
GOO::main(int argc, char *argv[])
{
  int i;

  SWIG_library_directory("goo");

  Preprocessor_define ((void *) "SWIGGOO", 0);

  for (i = 1; i < argc; i++) {
    if (argv[i]) {
      if (strcmp (argv[i], "-goostub") == 0) {
	if (argv[i + 1]) {
	  f_goo_stub = NewFile(argv[i + 1], (char *) "w");
	  Swig_mark_arg (i);
          Swig_mark_arg (i + 1);
	  i++;
	} else {
	  Swig_arg_error();
	}      
      } else if(strcmp (argv[i], "-goouse") == 0) {
	stub_to_goo_use = argv[i+1];
	Swig_mark_arg(i);
	Swig_mark_arg(i+1);
      } else if(strcmp (argv[i], "-goolibuse") == 0) {
	stub_to_lib_use = argv[i+1];
	Swig_mark_arg(i);
	Swig_mark_arg(i+1);
      }
    }
  }
  

  SWIG_config_file("goo.i");
}

int
GOO::top(Node *n)
{
  char init_name_buf[256];

  /* Initialize all of the output files */
  String *outfile = Getattr(n,"outfile");

  f_runtime = NewFile(outfile,"w");
  if (!f_runtime) {
    Printf(stderr,"*** Can't open '%s'\n", outfile);
    SWIG_exit(EXIT_FAILURE);
  }
  f_init = NewString("");
  f_header = NewString("");
  f_wrappers = NewString("");

#ifdef SPLAT_MODULE_INFO
  f_binding_info = NewString("");
  f_export_info = NewString("");
#endif

  /* Register file targets with the SWIG file handler */
  Swig_register_filebyname("header",f_header);
  Swig_register_filebyname("wrapper",f_wrappers);
  Swig_register_filebyname("runtime",f_runtime);
  Swig_register_filebyname("init",f_init);

  Printf(f_runtime, "/* -*- buffer-read-only: t -*- vi: set ro: */\n");
  Swig_banner (f_runtime);

  Printf (f_runtime, "/* Implementation : GOO */\n\n");

  Printf(f_runtime, "#include \"grt.h\"\n");
  //  Printf(f_runtime, "#include \"dlgrt.h\"\n\n");

  /* Write out directives and declarations */

  if (NoInclude) {
    Printf(f_runtime, "#define SWIG_NOINCLUDE\n");
  }
  
  module = Swig_copy_string(Char(Getattr(n,"name")));

  char module_name[256];

  if (!module)
    sprintf(module_name, "swig");
  else {
    strcpy(module_name,module);
  }

  if (CPlusPlus) {
    Printf(f_runtime, "extern \"C\" {\n\n");
  }

  make_goo_name(init_name_buf, module, "init_me");

#ifdef BUILD_DYNAMIC
  Printf (f_runtime, "extern void load_module_dl (void);\n");
#else
  Printf (f_runtime, "static void load_module_dl (void);\n");
#endif

#ifdef SPLAT_MODULE_INFO

  Printf (f_init, "extern MODULE_INFO *goo_toplevel_module_info;\n");
  Printf (f_init, "extern P YevalSastYreload_modules;\n");

#endif

  Printf (f_init, "P %s(void) {load_module_dl();}\n\n", init_name_buf);

#ifdef BUILD_DYNAMIC
  Printf (f_init, "extern void\nload_module_dl (void)\n{\n");
#else
  Printf (f_init, "static void\nload_module_dl (void)\n{\n");
#endif

  Printf (f_init, "DEFCREGS();\n");
  Printf (f_init, "  static int need_init = 1;\n  if(!need_init) return;\n\n");

#ifdef SPLAT_MODULE_INFO

  Printf (f_binding_info, "static BINDING_INFO binding_infos[] = {\n");

  Printf(f_export_info, "static EXPORT_INFO export_infos[] = {\n");

#endif

  if (CPlusPlus) {
    Printf(f_runtime, "\n}\n");
  }

  Language::top(n);
  
  /* Close module */

  Printf(f_wrappers,"#ifdef __cplusplus\nextern \"C\" {\n#endif\n");

  //  SwigType_emit_type_table (f_runtime, f_wrappers);

  Printf(f_wrappers,"#ifdef __cplusplus\n}\n#endif\n");

#ifdef SPLAT_MODULE_INFO

  Printf (f_init, " {\n  MODULE_INFO* old_info;\n");
  Printf (f_init, "  old_info = goo_toplevel_module_info;\n");
  Printf (f_init, "  goo_toplevel_module_info = &module_info_me;\n");
  Printf (f_init, "  XCALL0(1, YevalSastYreload_modules);\n");
  Printf (f_init, "  goo_toplevel_module_info = old_info;\n");
  Printf (f_init, " }\n");

#endif

  Printf (f_init, "}\n\n");

#ifdef SPLAT_MODULE_INFO
  
  Printf(f_binding_info, "{NULL, NVAR, NULL}\n};\n");
  Printf(f_export_info, "{NULL,NULL}\n};\n");

  Printf(f_export_info, "static USE_INFO nop_use_infos[] = {{NULL}};\n\n");
  Printf(f_export_info, "static IMPORT_INFO nop_import_infos[] = {{NULL,NULL,NULL}};\n\n");

  Printf(f_export_info, "static MODULE_INFO module_info_me = {\n");
  Printf(f_export_info, "  \"%s\",\n", module);
  Printf(f_export_info, "  NULL,\n");
  Printf(f_export_info, "  nop_use_infos, nop_import_infos,\n");
  Printf(f_export_info, "  binding_infos, export_infos\n};\n");

#endif

  if(f_goo_stub)
    {

      //(use samurui/gtk.swig.c)
  //(use/library gtk-x11-2.0)

      Printf(f_goo_stub, "(use goo)\n(use goo/boot)\n\n\n");
      if(stub_to_goo_use)
	{
	  Printf(f_goo_stub, "(use %s)\n", stub_to_goo_use);
	}
      if(stub_to_lib_use)
	{
	  Printf(f_goo_stub, "(use/library %s)\n", stub_to_lib_use);
	}

      // Put out all the dv's and exports.
      Printf(goo_text, ")\n");
      Printf(f_goo_stub, "%s", goo_text);

      // Put out our init dl, and then call it.
      Printf(f_goo_stub, "(dl init_me)\n");
      Printf(f_goo_stub, "(init_me)\n");
	      
      Delete(f_goo_stub);
      f_goo_stub = NULL;
    }
  else
    {
      printf("Warning! -goostub was not specified.\n");
    }

  /* Close all of the files */
  Dump(f_header,f_runtime);
  Dump(f_wrappers,f_runtime);

#ifdef SPLAT_MODULE_INFO
  Dump(f_binding_info,f_runtime);
  Dump(f_export_info,f_runtime);
#endif

  Wrapper_pretty_print(f_init,f_runtime);
  Delete(f_header);
  Delete(f_wrappers);
  Delete(f_init);

#ifdef SPLAT_MODULE_INFO
  Delete(f_binding_info);
  Delete(f_export_info);
#endif

  Close(f_runtime);
  Delete(f_runtime);
  return SWIG_OK;
}

inline char goo_map_char(char c)
{
  switch(c)
    {
    case '%':
      return 'P';
    case '_':
      return 'U';
    case '/':
      return 'S';
    default:
      return tolower(c);
    }
}

void goo_strcat(char *target, char *source)
{
  char *ip, *op;

  // Scan until we hit the end
  for(op=target; *op != '\0'; op++);

  for(ip=source; *ip != '\0'; ip++,op++)
    {
      *op = goo_map_char(*ip);
    }

  *op = '\0';  
}

void make_goo_name(char *target, char *module, char *name)
{
  char *ip, *op;

  strcpy(target, "Y");
  goo_strcat(target, module);
  strcat(target, "Y");
  goo_strcat(target, name);
}

int
GOO::functionWrapper(Node *n)
{
  char *iname = GetChar(n,"sym:name");
  SwigType *d = Getattr(n,"type");
  ParmList *l = Getattr(n,"parms");
  Parm *p;
  DOHString *proc_name = 0;
  char source[256], target[256], wname[256];
  Wrapper *f = NewWrapper();
  String *cleanup = NewString("");
  String *outarg = NewString("");
  String *returns = NewString("");
  String *argType = NewString("");
  int returns_list = 0;
  String *tmp = NewString("");
  String *genVarLiterals = NewString(""); // define and instantiate the literals (Var names)
  String *varLitRefList  = NewString(""); // list of references to the literals
  String *varTypeRefList = NewString(""); // list of type references for the vars
  String *tm;
  char *gooType;
  int i;
  int numargs = 0;
  int numreq = 0;
  int argsOutted = 0;

  // Make a wrapper name for this
  {
    String *n = Swig_name_mangle(iname);
    //strcpy(wname, Char(Swig_name_wrapper(n)));
    make_goo_name(wname, module, Char(n));
  }

  Printf(f->code, "LINK_STACK();\n");

  /* Emit locals etc. into f->code; figure out which args to ignore */
  emit_args (d, l, f);

  /* Attach the standard typemaps */
  emit_attach_parmmaps(l,f);

  /* Get number of required and total arguments */
  numargs = emit_num_arguments(l);
  numreq  = emit_num_required(l);

  // Prep the variable lists
  Printf(varLitRefList, "%d%c",  numargs, (numargs == 0) ? ' ' : ',');
  Printf(varTypeRefList, "%d%c", numargs, (numargs == 0) ? ' ' : ',');

  /* Declare return variable */
  Wrapper_add_local(f, "retval", "P retval = (P)YPib((P)0)");

  /* Get the output typemap so we can start generating documentation.  Don't
     worry, the returned string is saved as 'tmap:out' */

  Swig_typemap_lookup_new("out",n,"result",0);
  
  if ((tm = Getattr(n,"tmap:out:doc"))) {
    Printv(returns,tm,0);
  } else {
    String *s = SwigType_str(d,0);
    Chop(s);
    Printf(returns,"<%s>",s);
    Delete(s);
  }

  /* Open prototype and signature */
  Printf(f->def, "EXT(%s,\"%s\",\"%s\");\n", wname, module, iname);
  Printf(f->def, "FUNFOR(%s);\n\n", wname);

  Printv(f->def, "FUNCODEDEF(", wname,")\n{\n", 0);

#ifdef SPLAT_MODULE_INFO
  Printf(f_binding_info, "{\"%s\", CVAR, &%s},\n", iname, wname);
  Printf(f_export_info, "{\"%s\", \"%s\"},\n", iname, iname);
#endif

  dudes_so_far++;
  if (dudes_so_far >= 50)
    {
      Printf(goo_text, ")\n\n(seq\n");
      dudes_so_far = 0;
    }

  Printf(goo_text, "(dv %s #f)\n", iname);
  Printf(goo_text, "(export %s)\n\n", iname);

  // Code: link the stack
  //  Printf(f->code, "printf(\"about to link stack\\n\");\n");

  /* Now write code to extract the parameters */
  if(numargs > 0)
    Printf(f->def, "P ");

  for (i = 0, p = l; i < numargs; i++) {
    while (Getattr(p,"tmap:ignore")) {
      p = Getattr(p,"tmap:ignore:next");
    }

    SwigType *pt = Getattr(p,"type");
    String   *pn = Getattr(p,"name");
    String   *ln = Getattr(p,"lname");

    // Produce names of source and target
    sprintf(source,"%s_",Char(pn));
    sprintf(target,"%s", Char(ln));

    if (i!=0) Printf(f->def,", ");

    // Define the variable
    Printf(f->def, source);

    // init: Construct the literal init code
    Printf(genVarLiterals, "  P lit_arg_%d = YPPsym((P)\"%s\");\n", i, pn);
    Printf(varLitRefList, "LITREF(lit_arg_%d)%c", i, (i == (numargs - 1)) ? ' ' : ',');
    // - create the type reference
    // if it's an enum, it should be an int
    pt = SwigType_strip_qualifiers(pt);
    if(SwigType_istypedef(pt))
    {
      SwigType *tpt;
      tpt = SwigType_typedef_resolve_all(pt);
      Delete(pt);
      pt = tpt;
    }

    if(SwigType_isenum(pt))
      gooType = "int";
    else if(strcmp(Char(pt), "P") == 0)
      gooType = "any";
    else if(strncmp(Char(pt), "a().p.", 6) == 0) // array of pointer -> tup!
      gooType = "tup";
    else if(SwigType_isarray(pt))
      gooType = "tup";
    else if(SwigType_ispointer(pt))
      if(strcmp(Char(pt), "p.char") == 0)
	gooType = "str";
      else
	gooType = "loc";
    else
      gooType = Char(pt);

    if(strncmp(gooType, "unsigned ", 9) == 0)
      gooType = (char *)&(((char *)gooType)[9]);
    if(strcmp(gooType, "long") == 0)
      gooType = "int";
    if(strcmp(gooType, "short") == 0)
      gooType = "int";
    if(strcmp(gooType, "float") == 0)
      gooType = "flo";
    if(strcmp(gooType, "double") == 0)
      gooType = "flo";
    if(strcmp(gooType, "char") == 0)
      gooType = "int";

    // Let them override us.

    if ((tm = Getattr(p,"tmap:in:doc")))
    {
      Printv(argType,tm,0);
      gooType = Char(argType);
      //      printf("--------Got explicit type of %s\n", gooType);
    }

    //    printf("Thing is: #%s# #%s# %d\n", Char(pt), gooType, SwigType_isconst(pt));
    Printf(varTypeRefList, "VARREF(YL%sG)%c", gooType, (i == (numargs - 1)) ? ' ' : ',');

    // Unpack the arguments from the stack
    //    Printf(f->code, " printf(\"about to get arg %d\\n\");\n", i);
    Printf(f->code, "  ARG(%s, %d);\n", (char *)source, i);
    //    Printf(f->code, " printf(\"arg is %cd\\n\", %s);\n", '%', (char *)source);

    if ((tm = Getattr(p,"tmap:in"))) {
      Replace(tm,"$source",source,DOH_REPLACE_ANY); /* Deprecated */
      Replace(tm,"$target",target,DOH_REPLACE_ANY); /* Deprecated */
      Replace(tm,"$input",source,DOH_REPLACE_ANY);
      Setattr(p,"emit:input", source);
      Printv(f->code,tm,"\n",0);

      p = Getattr(p,"tmap:in:next");
    } else {
      // error...
      p = nextSibling(p);
    }
  }

  if(numargs > 0)
    Printf(f->def, ";");

  /* Insert constraint checking code */
  for (p = l; p;) {
    if ((tm = Getattr(p,"tmap:check"))) {
      Replace(tm,"$target",Getattr(p,"lname"),DOH_REPLACE_ANY);
      Printv(f->code,tm,"\n",0);
      p = Getattr(p,"tmap:check:next");
    } else {
      p = nextSibling(p);
    }
  }
  /* Pass output arguments back to the caller. */

  /* Insert argument output code */
  for (p = l; p;) {
    if ((tm = Getattr(p,"tmap:argout"))) {
      Replace(tm,"$source",Getattr(p,"lname"),DOH_REPLACE_ANY);   /* Deprecated */
      Replace(tm,"$target",Getattr(p,"lname"),DOH_REPLACE_ANY);   /* Deprecated */
      Replace(tm,"$arg",Getattr(p,"emit:input"), DOH_REPLACE_ANY);
      Replace(tm,"$input",Getattr(p,"emit:input"), DOH_REPLACE_ANY);
      Replace(tm,"$result", "retval", DOH_REPLACE_ANY);
      Printv(outarg,tm,"\n",0);
      argsOutted++;

      p = Getattr(p,"tmap:argout:next");
    } else {
      p = nextSibling(p);
    }
  }

  /* Insert cleanup code */
  for (p = l; p;) {
    if ((tm = Getattr(p,"tmap:freearg"))) {
      Replace(tm,"$target",Getattr(p,"lname"),DOH_REPLACE_ANY);
      Printv(cleanup,tm,"\n",0);
      p = Getattr(p,"tmap:freearg:next");
    } else {
      p = nextSibling(p);
    }
  }

  // Now write code to make the function call
  emit_action(n,f);

  // Now have return value, figure out what to do with it.
  if ((tm = Getattr(n,"tmap:out"))) {
    SwigType *d_stripped;

    Replaceall(tm,"$result","retval");
    Replaceall(tm,"$target","retval");
    Replaceall(tm,"$source","result");

    d_stripped  = SwigType_strip_qualifiers(d);
    
    if(SwigType_isarray(d_stripped))
      {
	SwigType_pop(d_stripped);

	if(SwigType_ispointer(d_stripped))
	  {
	    Replaceall(tm, "$opt_ref", ""); // no deref! it's an array of pointers
	  }
	else
	  {
	    Replaceall(tm, "$opt_ref", "&"); // ref! it's an array!
	  }
      }
    else
      {
	Replaceall(tm, "$opt_ref", ""); // no array, no pointers
      }

    Delete(d_stripped);

    Printv(f->code,tm,"\n",0);
  }
  else {
    // error
  }

  // Dump the argument output code
  Printv(f->code,outarg,0);

  // Dump the argument cleanup code
  Printv(f->code,cleanup,0);

  // Look for any remaining cleanup

  if ((NewObject) || (Getattr(n,"feature:new"))) {
    if ((tm = Swig_typemap_lookup_new("newfree",n,"result",0))) {
      Replaceall(tm,"$source","result");
      Printv(f->code,tm,"\n",0);
    }
  }

  // Free any memory allocated by the function being wrapped..
  if ((tm = Swig_typemap_lookup_new("ret",n,"result",0))) {
    Replaceall(tm,"$source","result");
    Printv(f->code,tm,"\n",0);
  }

  // Exit code

  // code: Restore Stack
  Printf(f->code, "UNLINK_STACK();\n");

  // return! yay!
  Printf(f->code, "QRET(retval);\n");

  // close it out?
  Printf(f->code, "}\n");

  Wrapper_print (f, f_wrappers);

  {
    Printf(f_init, " {\n");

    Printf(f_init, "  P methodSig, methodDef;\n");
    Printf(f_init, "  P methodName = YPPsym((P)\"%s\");\n", iname); 

    // Inject the literal instantiations
    Printf(f_init, Char(genVarLiterals));

    // Construct the method signature
    // void should be any!
    // THIS IS OFFICIALLY NOW HORRIBLE AND UGLY
    // it could probably also be done some cleaner way with typemaps
    //  too... but still, hacking this is somehow easier...
    // if you are some poor soul who is not me, you are probably better
    //  off just rewriting this using typemaps (if possible), or 
    //  writing a happy loop with conversion tables
    d = SwigType_strip_qualifiers(d);
    if(SwigType_istypedef(d))
    {
      SwigType *td;
      td = SwigType_typedef_resolve_all(d);
      Delete(d);
      d = td;
    }
       
    if(strcmp(Char(d), "void") == 0)
      gooType = "any";
    else if(SwigType_isenum(d))
      gooType = "int";
    else if(strcmp(Char(d), "P") == 0)
      gooType = "any";
    else if(strncmp(Char(d), "a().p.", 6) == 0) // array of pointer -> tup
      gooType = "tup";
    else if(SwigType_isarray(d))
      gooType = "tup";
    else if(SwigType_ispointer(d))
      if(strcmp(Char(d), "p.char") == 0)
	gooType = "str";
      else
	gooType = "loc";
    else
      gooType = Char(d);

    if(strncmp(gooType, "unsigned ", 9) == 0)
      gooType = (char *)&(((char *)gooType)[9]);
    if(strcmp(gooType, "long") == 0)
      gooType = "int";
    if(strcmp(gooType, "short") == 0)
      gooType = "int";
    if(strcmp(gooType, "float") == 0)
      gooType = "flo";
    if(strcmp(gooType, "double") == 0)
      gooType = "flo";
    if(strcmp(gooType, "char") == 0)
      gooType = "int";

    if(argsOutted > 0)
      gooType = "tup";

      //    printf("checking type of #%s# #%s#\n", Char(d), gooType);

    Printf(f_init, "  methodSig = YPfab_sig(YPPlist(%s), YPfalse, YPint((P)%d), VARREF(YL%sG), Ynil);\n", Char(varTypeRefList), numargs, gooType); // (d == NULL) ? "any" : SwigType_str(d, ""), numargs);

    // Define the method
    Printf(f_init, "  methodDef = YPfab_met(FUNCODEREF(%s), methodSig, LITREF(methodName), YPPlist(%s), YPfalse, YPfalse);\n", wname, Char(varLitRefList));

    // Register the method
    Printf(f_init, "  VARSET(%s, methodDef);\n", wname);

    Printf(f_init, " }\n");
  }

  Delete(outarg);
  Delete(cleanup);
  Delete(returns);
  Delete(argType);
  Delete(tmp);
  DelWrapper(f);
  return SWIG_OK;
}

int
GOO::variableWrapper(Node *n)
{
  return SWIG_OK;
}

int
GOO::constantWrapper(Node *n)
{
  char *name      = GetChar(n,"name");
  char *iname     = GetChar(n,"sym:name");
  SwigType *type  = Getattr(n,"type");
  String   *value = Getattr(n,"value");

  int OldReadOnly = ReadOnly;
  char   var_name[256];
  DOHString *rvalue;
  Wrapper *f;
  SwigType *nctype;
  String   *tm;

  f = NewWrapper();
  ReadOnly = 1;     // Enable readonly mode.

  // Make a static variable;

  make_goo_name(var_name, module, iname);
  //  sprintf (var_name, "const_%s", iname);

  // Strip const qualifier from type if present

  nctype = NewString(type);
  if (SwigType_isconst(nctype)) {
    Delete(SwigType_pop(nctype));
  }
  
  if ((SwigType_type(nctype) == T_USER)) {
    Printf (stderr, "%s : Line %d.  Unsupported constant value.\n",
             input_file, line_number);
    return SWIG_NOWRAP;
  }

  // See if there's a typemap

  if (SwigType_type(nctype) == T_STRING) {
    rvalue = NewStringf("\"%s\"", value);
  } else if (SwigType_type(nctype) == T_CHAR) {
    rvalue = NewStringf("\'%s\'", value);
  } else {
    rvalue = NewString(value);
  }

  if ((tm = Swig_typemap_lookup_new("constant",n,name,0))) {
    Replaceall(tm,"$source",rvalue);
    Replaceall(tm,"$value",rvalue);
    Replaceall(tm,"$target",name);
    Printv(f_header,tm,"\n",0);
  } else {
    // Create variable and assign it a value
    Printf (f_header, "EXT(%s,\"%s\",\"%s\");\n", var_name, module, iname);
    Printf (f_init, "  VARSET(%s,(P)YPib((P)%s));\n", var_name, value);

  dudes_so_far++;
  if (dudes_so_far >= 50)
    {
      Printf(goo_text, ")\n\n(seq\n");
      dudes_so_far = 0;
    }

    Printf(goo_text, "(dv %s #f)\n", iname);
    Printf(goo_text, "(export %s)\n\n", iname);

#ifdef SPLAT_MODULE_INFO
    Printf(f_binding_info, "{\"%s\", CVAR, &%s},\n", iname, var_name);
    Printf(f_export_info, "{\"%s\", \"%s\"},\n", iname, iname);
#endif
  }
  {
    /* Hack alert: will cleanup later -- Dave */
    Node *n = NewHash();
    Setattr(n,"name",var_name);
    Setattr(n,"sym:name",iname);
    Setattr(n,"type", nctype);
    variableWrapper(n);
    Delete(n);
  }
  ReadOnly = OldReadOnly;
  Delete(nctype);
  Delete(rvalue);
  DelWrapper(f);
  return SWIG_OK;
}

int
GOO::membervariableHandler(Node *n)
{
  /* Only emit traditional VAR-get and VAR-set procedures */
  Language::membervariableHandler(n);

  return SWIG_OK;
}







