/******************************************************************************
 * Simplified Wrapper and Interface Generator  (SWIG)
 *
 * Author : David Beazley
 *
 * Department of Computer Science
 * University of Chicago
 * 1100 E 58th Street
 * Chicago, IL  60637
 * beazley@cs.uchicago.edu
 *
 * Please read the file LICENSE for the copyright and terms by which SWIG
 * can be used and distributed.
 *****************************************************************************/

/* Scheme48 module for SWIG; the code is hopefully a relatively straight
    adaptation of the guile module code.  Anything which isn't is probably
    not done correctly ;)
*/

#include "mod11.h"
#include "scheme48.h"
#ifndef MACSWIG
#include "swigconfig.h"
#endif
#include <ctype.h>

static char *scheme48_usage = (char*)"\
   It's a secret to everybody.
\n";

static  File         *f_runtime = 0;
static  File         *f_header = 0;
static  File         *f_wrappers = 0;
static  File         *f_init = 0;

// ---------------------------------------------------------------------
// SCHEME48 ()
// ---------------------------------------------------------------------

SCHEME48::SCHEME48 ()
{
  // Set global vars

  SWIG_typemap_lang("scheme48");

  // Set class vars

  prefix = (char*)"";
  module = NULL;
  package = NULL;
  scmstub = NULL;
  emit_setters = 0;
  struct_member = 0;
  before_return = NULL;
  exported_symbols = NewString("");
  scmtext = NewString("");
  Swig_register_filebyname("scheme", scmtext);
}

// ---------------------------------------------------------------------
// SCHEME48::main()
//
// Parse arguments.
// ---------------------------------------------------------------------

void
SCHEME48::main (int argc, char *argv[])
{
  int i, orig_len;

  SWIG_library_directory("scheme48");

  // Look for certain command line options
  for (i = 1; i < argc; i++) {
    if (argv[i]) {
      if (strcmp (argv[i], "-help") == 0) {
	fputs (scheme48_usage, stderr);
	SWIG_exit (EXIT_SUCCESS);
      }
      else if (strcmp (argv[i], "-prefix") == 0) {
	if (argv[i + 1]) {
	  prefix = new char[strlen (argv[i + 1]) + 2];
	  strcpy (prefix, argv[i + 1]);
	  Swig_mark_arg (i);
	  Swig_mark_arg (i + 1);
	  i++;
	} else {
	  Swig_arg_error();
	}
      }
      else if (strcmp (argv[i], "-package") == 0) {
	if (argv[i + 1]) {
	  package = new char[strlen (argv[i + 1]) + 2];
	  strcpy (package, argv [i + 1]);
	  Swig_mark_arg (i);
	  Swig_mark_arg (i + 1);
	  i++;
	} else {
	  Swig_arg_error();
	}
      }
      else if (strcmp (argv[i], "-emit-setters") == 0) {
	emit_setters = 1;
	Swig_mark_arg (i);
      }
      else if (strcmp (argv[i], "-scmstub") == 0) {
	if (argv[i + 1]) {
	  scmstub = NewFile(argv[i + 1], (char *) "w");
	  Swig_mark_arg (i);
          Swig_mark_arg (i + 1);
	  i++;
	} else {
	  Swig_arg_error();
        }
      }
    }
  }

  // Make sure `prefix' ends in an underscore
  /*
  orig_len = strlen (prefix);
  if (prefix[orig_len - 1] != '_') {
    prefix[1 + orig_len] = 0;
    prefix[orig_len] = '_';
  }
  */

  /* Add a symbol for this module */
  Preprocessor_define ((void *) "SWIGSCHEME48",0);

  /* Read in default typemaps */
  SWIG_config_file("scheme48.i");
}

// --------------------------------------------------------------------
// SCHEME48::top()
// --------------------------------------------------------------------

int
SCHEME48::top(Node *n)
{

  /* Initialize all of the output files */
  String *outfile = Getattr(n,"outfile");

  f_runtime = NewFile(outfile,"w");
  if (!f_runtime) {
    Printf(stderr,"*** Can't open '%s'\n", outfile);
    SWIG_exit(EXIT_FAILURE);
  }
  f_init = NewString("");
  f_header = NewString("");
  f_wrappers = NewString("");

  /* Register file targets with the SWIG file handler */
  Swig_register_filebyname("header",f_header);
  Swig_register_filebyname("wrapper",f_wrappers);
  Swig_register_filebyname("runtime",f_runtime);
  Swig_register_filebyname("init",f_init);

  Printf(f_runtime, "/* -*- buffer-read-only: t -*- vi: set ro: */\n");
  Swig_banner (f_runtime);

  Printf (f_runtime, "/* Implementation : SCHEME48 */\n\n");
  Printf (f_runtime, "#include <scheme48.h>\n\n");

  /* Write out directives and declarations */

  if (NoInclude) {
    Printf(f_runtime, "#define SWIG_NOINCLUDE\n");
  }
  
  module = Swig_copy_string(Char(Getattr(n,"name")));

  if (CPlusPlus) {
    Printf(f_runtime, "extern \"C\" {\n\n");
  }

    /* Simple linkage; we have to export the SWIG_init function. The user can
       rename the function by a #define. */
    Printf (f_runtime, "extern void\nSWIG_init (void)\n;\n");
    Printf (f_init, "extern void\nSWIG_init (void)\n{\n");


  if (CPlusPlus) {
    Printf(f_runtime, "\n}\n");
  }

  Language::top(n);
  
  /* Close module */

  //  Printf(f_wrappers,"#ifdef __cplusplus\nextern \"C\" {\n#endif\n");

  //  SwigType_emit_type_table (f_runtime, f_wrappers);

  //  Printf(f_wrappers,"#ifdef __cplusplus\n}\n#endif\n");

  Printf (f_init, "}\n\n");
  char module_name[256];

  if (!module)
    sprintf(module_name, "swig");
  else {
    if (package)
      sprintf(module_name,"%s/%s", package,module);
    else
      strcpy(module_name,module);
  }
  emit_linkage (module_name);
  
  if (scmstub) {
    Delete(scmstub);
    scmstub = NULL;
  }

  /* Close all of the files */
  Dump(f_header,f_runtime);
  Dump(f_wrappers,f_runtime);
  Wrapper_pretty_print(f_init,f_runtime);
  Delete(f_header);
  Delete(f_wrappers);
  Delete(f_init);
  Close(f_runtime);
  Delete(f_runtime);
  return SWIG_OK;
}

void
SCHEME48::emit_linkage (char *module_name)
{
  DOHString *module_func = NewString("");

  if(scmstub == NULL)
    {
      printf("!!! Someone forgot to specify -scmstub!\n");
      return;
    }

  if (CPlusPlus) {
    Printf(f_init, "extern \"C\" {\n\n");
  }

  Printv(module_func,module_name,0);
  Replace(module_func,"-", "_", DOH_REPLACE_ANY);

    Printf (f_init, "\n/* Linkage: simple */\n");

    /* Emit Scheme stub if requested */
    DOHString *mod = NewString(module_name);
    Replace(mod, "/", " ", DOH_REPLACE_ANY);
    //    Printf (scmstub, ";;; -*- buffer-read-only: t -*- vi: set ro: */\n");
    Printf (scmstub, ";; You will want to paste the following into scheme48:\n");
    Printf (scmstub, ";; ,open dynamic-externals shared-bindings external-calls\n");
    //    Printf (scmstub, ";;; Automatically generated by SWIG; do not edit.\n\n");
    Delete(mod);

    Printf(scmstub, ";; You will probably want to correct the dynamic-load path to be\n");
    Printf(scmstub, ";;  absolute unless ldconfig knows about it...\n");
    Printf(scmstub, "(dynamic-load \"%s.so\")\n", module_name);
    Printf(scmstub, "(define zoop (get-external \"SWIG_init\"))\n");
    Printf(scmstub, "(call-external zoop)\n");

    Printf (scmstub, "%s", scmtext);

    /*
    if (Len(exported_symbols) > 0) {
      DOHString *ex = NewString(exported_symbols);
      Replace(ex, ", ", "\n        ", DOH_REPLACE_ANY);
      Replace(ex, "\"", "", DOH_REPLACE_ANY);
      Chop(ex);
      Printf(scmstub, "\n(export %s)\n", ex);
      Delete(ex);
    }
    */
  
  Delete(module_func);
  if (CPlusPlus) {
    Printf(f_init, "\n}\n");
  }
}

/* Return true iff T is a pointer type */

static int
is_a_pointer (SwigType *t)
{
  return SwigType_ispointer(SwigType_typedef_resolve_all(t));
}

/* Report an error handling the given type. */

static void
throw_unhandled_scheme48_type_error (SwigType *d)
{
  Printf (stderr, "%s : Line %d. Unable to handle type %s.\n",input_file, line_number, SwigType_str(d,0));
  error_count++;
}


// SCHEME48::functionWrapper()
// Create a function declaration and register it with the interpreter.

int
SCHEME48::functionWrapper(Node *n) {
  char *iname = GetChar(n,"sym:name");
  SwigType *d = Getattr(n,"type");
  ParmList *l = Getattr(n,"parms");
  Parm *p;
  DOHString *proc_name = 0;
  char source[256], target[256], wname[256];
  Wrapper *f = NewWrapper();;
  String *cleanup = NewString("");
  String *outarg = NewString("");
  String *signature = NewString("");
  String *returns = NewString("");
  int returns_list = 0;
  String *tmp = NewString("");
  String *tm;
  int i;
  int numargs = 0;
  int numreq = 0;

  // Make a wrapper name for this
  {
    String *n = Swig_name_mangle(iname);
    strcpy(wname, Char(Swig_name_wrapper(n)));
  }
  // Build the name for scheme.
  proc_name = NewString(iname);
  Replace(proc_name,"_", "-", DOH_REPLACE_ANY);

  if((Char(proc_name))[0] == '-')
    Delitem(proc_name, 0);

  /* Emit locals etc. into f->code; figure out which args to ignore */
  emit_args (d, l, f);

  /* Attach the standard typemaps */
  emit_attach_parmmaps(l,f);

  /* Get number of required and total arguments */
  numargs = emit_num_arguments(l);
  numreq  = emit_num_required(l);

  /* Declare return variable */

  Wrapper_add_local (f,"swig_result", "s48_value swig_result = 0");
  
  Wrapper_add_local(f, "argsout", "int argsout = 0");

  /* Get the output typemap so we can start generating documentation.  Don't
     worry, the returned string is saved as 'tmap:out' */

  Swig_typemap_lookup_new("out",n,"result",0);
  
  String *s = SwigType_str(d,0);
  Chop(s);
  Printf(returns,"<%s>",s);
  Delete(s);

  /* Open prototype and signature */

  Printv(f->def, "s48_value \n", wname," (", 0);
  Printv(signature, proc_name, 0);

  // start the import definition in the stub
  Printf(scmtext, "(import-lambda-definition %s (", proc_name); 

  /* Now write code to extract the parameters */

  for (i = 0, p = l; i < numargs; i++) {
    while (Getattr(p,"tmap:ignore")) {
      p = Getattr(p,"tmap:ignore:next");
    }

    SwigType *pt = Getattr(p,"type");
    String   *pn = Getattr(p,"name");
    String   *ln = Getattr(p,"lname");
    int opt_p = (i >= numreq);

    // Produce names of source and target
    sprintf(source,"s_%d",i);
    sprintf(target,"%s", Char(ln));

    if (i!=0)
      {
	Printf(f->def,", ");
	Printf(scmtext, " ");
      }

    Printf(f->def,"s48_value s_%d", i);

    Printf(scmtext, "%s", Char(pn));

    //    if (opt_p) {
    //      Printf(f->code,"    if (s_%d != GH_NOT_PASSED) {\n", i);
    //    }
    if ((tm = Getattr(p,"tmap:in"))) {
      Replace(tm,"$source",source,DOH_REPLACE_ANY); /* Deprecated */
      Replace(tm,"$target",target,DOH_REPLACE_ANY); /* Deprecated */
      Replace(tm,"$input",source,DOH_REPLACE_ANY);
      Setattr(p,"emit:input", source);
      Printv(f->code,tm,"\n",0);

      p = Getattr(p,"tmap:in:next");
    } else {
      throw_unhandled_scheme48_type_error (pt);
      p = nextSibling(p);
    }
    if (opt_p)
      Printf(f->code,"    }\n");
  }

  /* Insert constraint checking code */
  for (p = l; p;) {
    if ((tm = Getattr(p,"tmap:check"))) {
      Replace(tm,"$target",Getattr(p,"lname"),DOH_REPLACE_ANY);
      Printv(f->code,tm,"\n",0);
      p = Getattr(p,"tmap:check:next");
    } else {
      p = nextSibling(p);
    }
  }
  /* Pass output arguments back to the caller. */

  /* Insert argument output code */
  for (p = l; p;) {
    if ((tm = Getattr(p,"tmap:argout"))) {
      Replace(tm,"$source",Getattr(p,"lname"),DOH_REPLACE_ANY);   /* Deprecated */
      Replace(tm,"$target",Getattr(p,"lname"),DOH_REPLACE_ANY);   /* Deprecated */
      Replace(tm,"$arg",Getattr(p,"emit:input"), DOH_REPLACE_ANY);
      Replace(tm,"$input",Getattr(p,"emit:input"), DOH_REPLACE_ANY);
      Replace(tm,"$result", "swig_result", DOH_REPLACE_ANY);
      Printv(outarg,tm,"\n",0);
      p = Getattr(p,"tmap:argout:next");
    } else {
      p = nextSibling(p);
    }
  }

  /* Insert cleanup code */
  for (p = l; p;) {
    if ((tm = Getattr(p,"tmap:freearg"))) {
      Replace(tm,"$target",Getattr(p,"lname"),DOH_REPLACE_ANY);
      Printv(cleanup,tm,"\n",0);
      p = Getattr(p,"tmap:freearg:next");
    } else {
      p = nextSibling(p);
    }
  }

  /* Close prototype */

  Printf(f->def, ")\n{\n");

  Printf(scmtext, ") \"%s\")\n", proc_name);

  Printv(f->def, "#define FUNC_NAME \"", proc_name, "\"", 0);

  // Now write code to make the function call
  //Printv(f->code, tab4, "gh_defer_ints();\n", 0);
  emit_action(n,f);
  //Printv(f->code, tab4, "gh_allow_ints();\n", 0);

  // Now have return value, figure out what to do with it.

  if ((tm = Getattr(n,"tmap:out"))) {
    SwigType *d_stripped;

    Replaceall(tm,"$result","swig_result");
    Replaceall(tm,"$target","swig_result");
    Replaceall(tm,"$source","result");

    d_stripped  = SwigType_strip_qualifiers(d);
    
    if(SwigType_isarray(d_stripped))
      {
	SwigType_pop(d_stripped);

	if(SwigType_ispointer(d_stripped))
	  {
	    Replaceall(tm, "$opt_ref", ""); // no deref! it's an array of pointers
	  }
	else
	  {
	    Replaceall(tm, "$opt_ref", "&"); // ref! it's an array!
	  }
      }
    else
      {
	Replaceall(tm, "$opt_ref", ""); // no array, no pointers
      }

    Delete(d_stripped);

    Printv(f->code,tm,"\n",0);
  }
  else {
    throw_unhandled_scheme48_type_error (d);
  }

  // Dump the argument output code
  Printv(f->code,outarg,0);

  // Dump the argument cleanup code
  Printv(f->code,cleanup,0);

  // Look for any remaining cleanup

  if ((NewObject) || (Getattr(n,"feature:new"))) {
    if ((tm = Swig_typemap_lookup_new("newfree",n,"result",0))) {
      Replaceall(tm,"$source","result");
      Printv(f->code,tm,"\n",0);
    }
  }

  // Free any memory allocated by the function being wrapped..
  if ((tm = Swig_typemap_lookup_new("ret",n,"result",0))) {
    Replaceall(tm,"$source","result");
    Printv(f->code,tm,"\n",0);
  }

  // Wrap things up (in a manner of speaking)

  if (before_return)
    Printv(f->code, before_return, "\n", 0);
  Printv(f->code, "return swig_result;\n", 0);

  // Undefine the scheme name

  Printf(f->code, "#undef FUNC_NAME\n");
  Printf(f->code, "}\n");

  Wrapper_print (f, f_wrappers);

  /*
  if (emit_setters && struct_member && strlen(Char(proc_name))>3) {
    int len = Len(proc_name);
    const char *pc = Char(proc_name);
    // MEMBER-set and MEMBER-get functions.
    int is_setter = (pc[len - 3] == 's');
    if (is_setter) {
      Printf(f_init, "SCM setter = ");
      struct_member = 2; // have a setter
    }
    else Printf(f_init, "SCM getter = ");
    Printf (f_init, "gh_new_procedure(\"%s\", (swig_scheme48_proc) %s, %d, %d, 0);\n",
	    proc_name, wname, numreq, numargs-numreq);
    if (!is_setter) {
      // Strip off "-get"
      char *pws_name = (char*) malloc(sizeof(char) * (len - 3));
      strncpy(pws_name, pc, len - 3);
      pws_name[len - 4] = 0;
      if (struct_member==2) { 
	// There was a setter, so create a procedure with setter
	Printf (f_init, "gh_define(\"%s\", "
		"scm_make_procedure_with_setter(getter, setter));\n",
		pws_name);
      }
      else {
	// There was no setter, so make an alias to the getter
	Printf (f_init, "gh_define(\"%s\", getter);\n",
		pws_name);
      }
      Printf (exported_symbols, "\"%s\", ", pws_name);
      free(pws_name);
    }
  }
  else
*/
 {
    /* Register the function */
    Printf (f_init, "s48_define_exported_binding(\"%s\", s48_enter_pointer(%s));\n", proc_name, wname);
  }

  Printf (exported_symbols, "\"%s\", ", proc_name);

  Delete(proc_name);
  Delete(outarg);
  Delete(cleanup);
  Delete(signature);
  Delete(returns);
  Delete(tmp);
  DelWrapper(f);
  return SWIG_OK;
}

// -----------------------------------------------------------------------
// SCHEME48::variableWrapper()
//
// Create a link to a C variable.
// This creates a single function PREFIX_var_VARNAME().
// This function takes a single optional argument.   If supplied, it means
// we are setting this variable to some value.  If omitted, it means we are
// simply evaluating this variable.  Either way, we return the variables
// value.
// -----------------------------------------------------------------------

int
SCHEME48::variableWrapper(Node *n)
{
  // The scheme48 binding already is so superior to the Goo binding that 
   // I feel I would be doing Goo too great a dis-service by implementing
   // this for Scheme48. coughCOPOUTcough coughCOPOUTcough

  /*
  char *name  = GetChar(n,"name");
  char *iname = GetChar(n,"sym:name");
  SwigType *t = Getattr(n,"type");

  DOHString *proc_name;
  char  var_name[256];
  Wrapper *f;
  String  *tm;

  f = NewWrapper();
  // evaluation function names

  strcpy(var_name, Char(Swig_name_wrapper(iname))); 

  // Build the name for scheme.
  proc_name = NewString(iname);
  Replace(proc_name,"_", "-",DOH_REPLACE_ANY);

   if (1 || (SwigType_type(t) != T_USER) || (is_a_pointer(t))) {

    Printf (f->def, "static SCM\n%s(SCM s_0)\n{\n", var_name);

    // Define the scheme name in C.
    Printv(f->def, "#define FUNC_NAME \"", proc_name, "\"", 0);

    Wrapper_add_local (f, "swig_result", "s48_value swig_result");

    if (!ReadOnly) {
      // Check for a setting of the variable value
      Printf (f->code, "if (s_0 != GH_NOT_PASSED) {\n");
      if ((tm = Swig_typemap_lookup_new("varin",n,name,0))) {
	Replaceall(tm,"$source","s_0");
	Replaceall(tm,"$input","s_0");
	Replaceall(tm,"$target",name);
	Printv(f->code,tm,"\n",0);
      }
      else {
	throw_unhandled_scheme48_type_error (t);
      }
      Printf (f->code, "}\n");
    }

    // Now return the value of the variable (regardless
    // of evaluating or setting)

    if ((tm = Swig_typemap_lookup_new("varout",n,name,0))) {
      Replaceall(tm,"$source",name);
      Replaceall(tm,"$target","gswig_result");
      Replaceall(tm,"$result", "gswig_result");
      Printv(f->code,tm,"\n",0);
    }
    else {
      throw_unhandled_scheme48_type_error (t);
    }
    Printf (f->code, "\nreturn swig_result;\n");
    Printf (f->code, "#undef FUNC_NAME\n");
    Printf (f->code, "}\n");

    Wrapper_print (f, f_wrappers);

    // Now add symbol to the Scheme48 interpreter

    if (!emit_setters
	|| ReadOnly) {
      // Read-only variables become a simple procedure returning the
	//  value; read-write variables become a simple procedure with
	//  an optional argument.
      Printf (f_init, "\t gh_new_procedure(\"%s\", (swig_scheme48_proc) %s, 0, %d, 0);\n",
	      proc_name, var_name, ReadOnly ? 0 : 1);
    }
    else {
  // Read/write variables become a procedure with setter.
      Printf (f_init, "\t{ SCM p = gh_new_procedure(\"%s\", (swig_scheme48_proc) %s, 0, 1, 0);\n",
	      proc_name, var_name);
      Printf (f_init, "\t  gh_define(\"%s\", "
	      "scm_make_procedure_with_setter(p, p)); }\n",
	      proc_name);
    }
    Printf (exported_symbols, "\"%s\", ", proc_name);

  } else {
    Printf (stderr, "%s : Line %d. ** Warning. Unable to link with "
             " type %s (ignored).\n",
             input_file, line_number, SwigType_str(t,0));
  }
  Delete(proc_name);
  DelWrapper(f);
*/
  return SWIG_OK;
}

// -----------------------------------------------------------------------
// SCHEME48::constantWrapper()
//
// We create a read-only variable.
// ------------------------------------------------------------------------

int
SCHEME48::constantWrapper(Node *n)
{
  
  char *name      = GetChar(n,"name");
  char *iname     = GetChar(n,"sym:name");
  SwigType *type  = Getattr(n,"type");
  String   *value = Getattr(n,"value");

  int OldReadOnly = ReadOnly;
  DOHString *proc_name;
  char   var_name[256];
  char usable_value[1024];
  DOHString *rvalue;
  Wrapper *f;
  SwigType *nctype;
  String   *tm;

  f = NewWrapper();
  ReadOnly = 1;     // Enable readonly mode.

  // Make a static variable;

  if((prefix != NULL) && (strlen(prefix) > 0))
    {
      sprintf (var_name, "%s_%s", prefix, iname);
    }
  else
    {
      sprintf(var_name, "%s", iname);
    }

  // set usable_value to be, well, usable
  //  (namely, we want strings to be enclosed in double quotes!)
  if(strcmp("p.char", Char(type)) == 0)
    {
      sprintf(usable_value, "\"%s\"", Char(value));
    }
  else
    {
      sprintf(usable_value, "%s", Char(value));
    }

  // UPDATE!!! I forgot about enumerated values... so the code goes active again
  // The following code is commented out because although it is mind-bogglingly
   // brilliant, Scheme48's use of independent binding tables for scheme and
   // C makes it somewhat stupid and circumspect to publish the constants via
   // the C interface, just to have the scheme side pull them out and make 
   // them usable.  Now some people might thing we should provide getters
   // and setters, but those people are morons because these are constants.
   // Clearly, the brilliant thing to do is to just define the bloody constants
   // in the scheme file. That code follows this big-block-o-commented-out-code
  // Strip const qualifier from type if present

  nctype = NewString(type);
  if (SwigType_isconst(nctype)) {
    Delete(SwigType_pop(nctype));
  }
  
  // Build the name for scheme.
  proc_name = NewString(iname);
  Replace(proc_name,"_", "-", DOH_REPLACE_ANY);

  if ((SwigType_type(nctype) == T_USER) && (!is_a_pointer(nctype))) {
    Printf (stderr, "%s : Line %d.  Unsupported constant value.\n",
             input_file, line_number);
    return SWIG_NOWRAP;
  }

  // See if there's a typemap

  if (SwigType_type(nctype) == T_STRING) {
    rvalue = NewStringf("\"%s\"", value);
  } else if (SwigType_type(nctype) == T_CHAR) {
    rvalue = NewStringf("\'%s\'", value);
  } else {
    rvalue = NewString(value);
  }

  // attempt to get tmap:out ...
  Swig_typemap_lookup_new("out",n,usable_value,0);

  if ((tm = Swig_typemap_lookup_new("constant",n,name,0))) {
    Replaceall(tm,"$source",rvalue);
    Replaceall(tm,"$value",rvalue);
    Replaceall(tm,"$target",name);
    Printv(f_header,tm,"\n",0);
  } 
  else if ((tm = Getattr(n,"tmap:out"))) {
    Printf(f_init, "{\n");
    Printf(f_init, "s48_value swig_result;\n");
    Replaceall(tm,"$result","swig_result");
    Replaceall(tm,"$target","swig_result");
    Replaceall(tm,"$source",usable_value);
    Printv(f_init,tm,"\n",0);
    Printf(f_init, "s48_define_exported_binding(\"%s\", swig_result);\n", var_name);
    Printf(f_init,"}\n");
  }
  else
    {
    // Assume it's an integer...
    Printf (f_init, "s48_define_exported_binding(\"%s\", s48_enter_integer(%s));\n", var_name, usable_value);
  }

  Printf(scmtext, "(define %s (shared-binding-ref (lookup-imported-binding \"%s\")))\n", var_name, var_name);
  // now save us some memory by killing the binding...
  Printf(scmtext, "(undefine-imported-binding \"%s\")\n", var_name);

  ReadOnly = OldReadOnly;
  //  Delete(nctype);
  //  Delete(proc_name);
  //  Delete(rvalue);
  DelWrapper(f);
  return SWIG_OK;
}

int SCHEME48::membervariableHandler(Node *n)
{
  if (emit_setters) {
    struct_member = 1;
    Printf(f_init, "{\n");
    Language::membervariableHandler(n);
    Printf(f_init, "}\n");
    struct_member = 0;
  }
  else {
    /* Only emit traditional VAR-get and VAR-set procedures */
    Language::membervariableHandler(n);
  }
  return SWIG_OK;
}

void SCHEME48::pragma(char *lang, char *cmd, char *value)
{
  if (strcmp(lang,(char*)"scheme48") == 0) {
    if (strcmp(cmd, (char*)"beforereturn")==0) {
      if (before_return)
	Delete(before_return);
      before_return = value ? NewString(value) : NULL;
    }
  }
}


int SCHEME48::validIdentifier(String *s) {
  char *c = Char(s);
  /* Check whether we have an R5RS identifier.*/
  /* <identifier> --> <initial> <subsequent>* | <peculiar identifier> */
  /* <initial> --> <letter> | <special initial> */
  if (!(isalpha(*c) || (*c == '!') || (*c == '$') || (*c == '%')
	|| (*c == '&') || (*c == '*') || (*c == '/') || (*c == ':')
	|| (*c == '<') || (*c == '=') || (*c == '>') || (*c == '?')
	|| (*c == '^') || (*c == '_') || (*c == '~'))) {
    /* <peculiar identifier> --> + | - | ... */
    if ((strcmp(c, "+") == 0)
	|| strcmp(c, "-") == 0
	|| strcmp(c, "...") == 0) return 1;
    else return 0;
  }
  /* <subsequent> --> <initial> | <digit> | <special subsequent> */
  while (*c) {
    if (!(isalnum(*c) || (*c == '!') || (*c == '$') || (*c == '%')
	  || (*c == '&') || (*c == '*') || (*c == '/') || (*c == ':')
	  || (*c == '<') || (*c == '=') || (*c == '>') || (*c == '?')
	  || (*c == '^') || (*c == '_') || (*c == '~') || (*c == '+')
	  || (*c == '-') || (*c == '.') || (*c == '@'))) return 0;
    c++;
  }
  return 1;
}
